# -*- coding: utf-8 -*-
## Greedy Algorithm for NIRB
## Elise Grosjean
## 01/2021


from BasicTools.FE import FETools as FT
import numpy as np
from scipy import linalg
import sys
import os
import os.path as osp

import Readers as MR

from BasicTools.FE import FETools as FT
#import pickle
#import SVD

from scipy import linalg
from scipy import interpolate
from scipy.interpolate import griddata 
from scipy.interpolate import interp1d #time interpolation

############################################################
############################
### GREEDY FUNCTIONS  ######
############################
############################################################

#Gram-Schmidt orthogonalization of basis_vectors w.r.t. snapshotCorrelationOperator norm
def orthoGS(basis_vectors,cptmodes,m,snapshotCorrelationOperator): 
    residuals2=np.array(np.reshape(basis_vectors,((cptmodes,m))))
    basis_vectorsVec=np.reshape(basis_vectors,((cptmodes,m)))
            
    residuals2[-1,:]=basis_vectorsVec[-1]-sum((residuals2[k,:]*np.dot(snapshotCorrelationOperator.dot(residuals2[k]),basis_vectorsVec[-1]) for k in range(residuals2.shape[0]-1)))            
        
    basis_vectors[-1]=np.reshape(residuals2[-1,:],((m,1)))
    return basis_vectors


#check orthogonality of matrix wrt CorrelationMatrix, orthonormality of orthoN==True
def orthogonality_check(Matrix,CorrelationMatrix,orthoN=False):
    """
    This function check for the pairwise orthogonality of the new basis
    """
    list_ = list(Matrix)
    dot_matrix = np.array([[np.dot(CorrelationMatrix.dot(item1), item2.T) for item1 in list_] for item2 in list_])
    A=dot_matrix

    if(orthoN==False):
        A=np.diag(np.diag(dot_matrix))
    else:
        A= np.eye(dot_matrix.shape[0])

    if (dot_matrix - A < 1e-12).all():
        print(True)
        return True
    else:
        error = dot_matrix - A
     
        print("max error with identity: ",np.max(np.abs(error)))
        return False
############################################################
#Greedy_algorithm:
#      _ snapshots: list (number of training parameters) of snapshots arrays (of shape number of time steps)
#      _ snapshotCorrelationOperator: L2 or H1 sparse matrix
#      _ h1ScalarProducMatrix if H1 orthogonalization
#      _ Tolerance for greedy
#      _ Otherwise a-priori number of modes
############################################################
def greedy_algorithm(snapshots,NumerOfTimesSteps,snapshotCorrelationOperator,h1ScalarProducMatrix, tolerance, NumberOfModes):
    """
    Compute the reduced basis using a greedy algorithm.
    """
   
    TolGreedy=1e-3 #Tolerance on the time steps
    DegreesOfFreedom=np.shape(snapshotCorrelationOperator)[0]
    NbParam=len(snapshots)
    
    # initialization: First parameter
    ind=0 #first parameter (orthonormalized, so norm(snapshot) > 1e-10)

    basis_vectors=[] # basis functions list
    snapshotsList=snapshots.copy()
    snapshotsArray=snapshotsList.pop(ind) #snapshots with the parameter ind...
    
    residuals=np.array(snapshotsArray) #...all time steps for first parameter 
    n,m = residuals.shape
            
    tol=1
    cptmodes=1 #First mode
    cpt=0
    
   
    tolold=tol
    matVecProduct0=snapshotCorrelationOperator.dot(residuals.T)

    # FIRST PARAMETER SELECTION //LOOP OVER TIME
    TimeIndices=[]
    while(tol >= TolGreedy and NumberOfModes+1> cptmodes):
   
    #while(3 >=  cptmodes and NumberOfModes+1> cptmodes): #3 is a priori given
        
        matVecProduct= snapshotCorrelationOperator.dot(residuals.T)
        NormTestMax=np.sqrt(np.diag(np.dot(residuals,matVecProduct)))
        TimeIndex=np.argmax(NormTestMax)
          
        mask = (TimeIndex not in TimeIndices)
        if mask==False: #check if not already in the time steps 
                
            TimeIndexOld=TimeIndex
            filtered_list = [elt for indx,elt in enumerate(list(np.arange(len(NormTestMax)))) if elt not in TimeIndices]
            TimeIndex = filtered_list[np.argmax(NormTestMax[filtered_list])] 
            print("New time index: ", TimeIndex)

        TimeIndices.append(TimeIndex)
        tol=NormTestMax[TimeIndex]

        basis_vectors.append(np.reshape(residuals[TimeIndex,:],((m,1)))/tol) # time index, added in the modes
        
        print(cptmodes," / ",NumberOfModes, " (same parameter loop, new time step)")
        print(NormTestMax[TimeIndex],"/",TolGreedy ,"(threshold)")
  
       
        residuals -= np.outer(basis_vectors[-1],np.dot(basis_vectors[-1].T,matVecProduct0)).T #update
        # Gram-Schmidt
        orthogonality=orthogonality_check(np.reshape(basis_vectors,((cptmodes,m))),snapshotCorrelationOperator)
        orthocpt=0
        while(10>=orthocpt and orthogonality==False):
            basis_vectors=orthoGS(basis_vectors,cptmodes,m,snapshotCorrelationOperator)
    
            orthogonality=orthogonality_check(np.reshape(basis_vectors,((cptmodes,m))),snapshotCorrelationOperator)
            orthocpt+=1

        cptmodes+=1
        cpt+=1
        """
        if(tol > tolold):
            print(tol)
            break;
        """
        tolold=tol
        print(cptmodes)
      
        
    # PARAMETER SELECTION
    cptParam=1
    while(NumberOfModes>cptmodes):
        matrix=np.array(snapshotsList)    
        residuals = np.reshape(matrix,(((NbParam-cptParam)*n,m)))
        
        matVecProduct0 = snapshotCorrelationOperator.dot(residuals.T)
        for i in range(len(basis_vectors)):
            residuals -= np.outer(basis_vectors[i],np.dot(basis_vectors[i].T,matVecProduct0)).T
           
        matVecProduct = snapshotCorrelationOperator.dot(residuals.T)
        NormTestMax = np.sqrt(np.diag(np.dot(residuals,matVecProduct)))
        Index = np.argmax(NormTestMax)
        
        IndexParam = int(Index/n+0.00001)
        print("New parameter: ",IndexParam)
        tol =np.max(NormTestMax)
    
        basis_vectors.append(np.reshape(residuals[Index,:]/tol,((m,1)))) #Update
        
        print(cptmodes," / ",NumberOfModes, "(new parameter chosen)")
        print(tol,"/",TolGreedy,"(threshold)")
        
        if (cptParam!=1):
            cptmodes+=1

        #Gram-Schmidt
        orthogonality=orthogonality_check(np.reshape(basis_vectors,((cptmodes,m))),snapshotCorrelationOperator)
        orthocpt=0
        while(10>=orthocpt and orthogonality==False):
            basis_vectors=orthoGS(basis_vectors,cptmodes,m,snapshotCorrelationOperator)
          
            orthogonality=orthogonality_check(np.reshape(basis_vectors,((cptmodes,m))),snapshotCorrelationOperator)
            orthocpt+=1

            
        snapshotsArray=snapshotsList.pop(IndexParam)
    
        residuals=np.array(snapshotsArray)
        cpt=0
        #tol=1
        oldtol=tol
        TimeIndices=[]
        # TIME STEPS SELECTION
        TimeIndices.append(int(Index/m+0.00001))
        while(2>=cpt and NumberOfModes > cptmodes): #greedy on time steps a priori cpt given
         
            if cpt==0:
                matVecProduct0=snapshotCorrelationOperator.dot(residuals.T)
                for i in range(len(basis_vectors)):
                    residuals -= np.outer(basis_vectors[i],np.dot(basis_vectors[i].T,matVecProduct0)).T
                   
            matVecProduct = snapshotCorrelationOperator.dot(residuals.T)
            NormTestMax = np.sqrt(np.diag(np.dot(residuals,matVecProduct)))
            TimeIndex = np.argmax(NormTestMax)
            mask = (TimeIndex not in TimeIndices)
            if mask==False: #check if not already in the time steps 
                
                TimeIndexOld=TimeIndex
                filtered_list = [elt for indx,elt in enumerate(list(np.arange(len(NormTestMax)))) if elt not in TimeIndices]
                TimeIndex = filtered_list[np.argmax(NormTestMax[filtered_list])]
                print("New time index: ", TimeIndex)
                #break
                
            print("time: ",TimeIndex)
            TimeIndices.append(TimeIndex)
            tol = np.max(NormTestMax)
            #if(tol >= oldtol-1e-8):
            #    break;
            basis_vectors.append(np.reshape(residuals[TimeIndex,:]/tol,((m,1))))
            cptmodes+=1
            print(cptmodes," / ",NumberOfModes, "(same parameter, new time step)")
            print(tol,"/",TolGreedy,"(threshold)")
            
            #Gram-Schmidt
            orthogonality=orthogonality_check(np.reshape(basis_vectors,((cptmodes,m))),snapshotCorrelationOperator)
            orthocpt=0
            
            while(10>=orthocpt and orthogonality==False):
                basis_vectors=orthoGS(basis_vectors,cptmodes,m,snapshotCorrelationOperator)
       
                orthogonality=orthogonality_check(np.reshape(basis_vectors,((cptmodes,m))),snapshotCorrelationOperator)
                orthocpt+=1
            
            residuals -= np.outer(basis_vectors[-1],np.dot(basis_vectors[-1].T,matVecProduct0)).T
            cpt+=1
            """
            if(tol >= tolold-1e-8):
                #cptParam+=1
                print(tol)
                break;
            """
            tolold=tol
            
            #if(tol >= oldtol-1e-8):
            #    break;
        cptParam+=1

    orthocpt=0
    orthogonality=orthogonality_check(np.reshape(basis_vectors,((cptmodes,m))),snapshotCorrelationOperator)
    while(10>=orthocpt and orthogonality==False):
        basis_vectors=orthoGS(basis_vectors,cptmodes,m,snapshotCorrelationOperator)
        
        orthogonality=orthogonality_check(np.reshape(basis_vectors,((cptmodes,m))),snapshotCorrelationOperator)
        orthocpt+=1


    reducedOrderBasisU=np.array(np.reshape(basis_vectors,((NumberOfModes,m))))
    
    cpt=0
    for i in range(NumberOfModes):
        reducedOrderBasisNorm=np.sqrt(reducedOrderBasisU[i,:]@(snapshotCorrelationOperator@reducedOrderBasisU[i,:]))
        reducedOrderBasisU[i,:]/=reducedOrderBasisNorm #np.sqrt(M[i,i]) #L2 orthonormalization
        
    RB=[]
    for k in range(NumberOfModes,NumberOfModes+1): #for k in range(1,NumberOfModes+1):
        
 
        reducedOrderBasisUp=reducedOrderBasisU[0:k,:]#[idx,:]
        
        orthogonality=orthogonality_check(reducedOrderBasisUp,snapshotCorrelationOperator,True)
        print("ortho l2?", orthogonality)
        
        RB.append(reducedOrderBasisUp)
    return RB



############################################################
############################
### MAIN ###################
############################
############################################################

############################################################
"""          Initialization                              """

############################################################
############################################################
TBeg=0. # starting time
TFend=1. #final time 
onlineParameter=sys.argv[1]

## Directories
currentFolder=os.getcwd()
dataFolder=currentFolder+"/TestT0a1"
FinedataFolder=osp.join(dataFolder,'FineSnapshots/'+sys.argv[2]) #for fine snapshotsS
CoarsedataFolder=osp.join(dataFolder,'CoarseSnapshots/'+sys.argv[3]+"/"+sys.argv[4]) #for coarse snapshots
print("fine folder: ", FinedataFolder)
print("coarse folder: ", CoarsedataFolder)


ns=0 #number of snapshots
count1=0
for _, folder, _ in os.walk(FinedataFolder): #number of folders in FineData
    count1 += len(folder)

ns=count1-1 #-1 because of the online parameter not included in offline snapshots
print("Number of snapshots: ",ns)
#ns=18

nev=int(sys.argv[5])   #nombre de modes
print("number of modes: ",nev)

time=0.0 #init
dimension=2 #2D
           
TF=int(len([name for name in os.listdir(FinedataFolder+"/1/")]))
print("Number of fine time steps: ",TF)

TG=int(len([name for name in os.listdir(CoarsedataFolder+"/1/")]))
print("Number of coarse time steps: ",TG)

dtF=float(sys.argv[2]) #fine time steps
dtG=float(sys.argv[4]) #coarse time steps

for time in np.arange(0, TFend+0.0001, dtF):
    if time>=TBeg: #.9999: #NIRB sur t0=]1, 2]
        t0f=time
        break
for time in np.arange(0, TFend+0.0001, dtG):
    if time>=TBeg: #.9999:
        t0g=time
        break
    
#for time interpolation
oldtime=np.arange(t0g, TFend+0.0001, dtG)
newtime=np.arange(t0f, TFend+0.0001, dtF)

#print(np.shape(oldtime),len(newtime)," ", TF," ",TG)

"""
-------------------
###  Read fine mesh
------------------- 
"""
meshFileName = FinedataFolder + "/1/Snapshoth_1.vtu";
mesh=MR.Readmesh(meshFileName)
mesh.nodes= mesh.nodes[:,:2] #2D

print("Fine mesh defined in " + meshFileName + " has been read")
nbeOfComponentsPrimal = 1 # 1 field 
numberOfNodes = mesh.GetNumberOfNodes()
print("DoF fine mesh ", numberOfNodes)
"""
-------------------
###  Read coarse mesh
------------------- 
"""

meshFileName2 = CoarsedataFolder + "/1/Snapshoth_1.vtu";#//FineMesh/mesh1.msh"
mesh2=MR.Readmesh(meshFileName2)
mesh2.nodes = mesh2.nodes[:,:2] #CAS 2D

print("Coarse mesh defined in " + meshFileName2 + " has been read")
numberOfNodes2 = mesh2.GetNumberOfNodes()
print("DoF coarse mesh ", numberOfNodes2)

"""
-------------------
###  mesh space interpolation
------------------- 
"""

inputnodes=mesh2.nodes
outputnodes=mesh.nodes

"""
-------------------
###  read all snapshots ...
------------------- 
"""

parameters=[]
for i in range(1,ns+2):
    if(float(0.5*i)%1>1e-3):
        parameters.append(str(0.5*i))
    else:
        parameters.append(str(int(0.5*i)))
parameters.remove(onlineParameter) #online parameter mu=1
print("parameters for heat equation:",parameters)

snapshots=[]

for e,i in enumerate(parameters):
    snapshotsTime=[]
    for time in range(0,TF):   
        snapshot =MR.VTKReadToNp("Velocity",FinedataFolder+"/"+i+"/Snapshoth_",time)
        snapshotsTime.append(snapshot)
    snapshots.append(snapshotsTime)

    
snapshotsH=[]

for e,i in enumerate(parameters):
    snapshotsHTime=[]
    for time in range(0,TG):

        snapshotH =MR.VTKReadToNp("Velocity",CoarsedataFolder+"/"+i+"/Snapshoth_",time)

        #Compute the projected data using the projection operator
        snapshotHSpaceinterpolated = griddata(inputnodes,snapshotH,outputnodes,method='linear')#//Operator.dot(snapshotH)
        snapshotsHTime.append(snapshotHSpaceinterpolated)
   
    interp  = interp1d(oldtime,snapshotsHTime,kind='quadratic',axis=0,fill_value="extrapolate")

    solutionUHI=interp(newtime) #time and space interpolation
    snapshotsH.append(solutionUHI)
  
snapshotsHarraytl=np.array(snapshotsH)

############################################################
"""          Greedy                                      """
############################################################

print("ComputeL2ScalarProducMatrix ...")

l2ScalarProducMatrix = FT.ComputeL2ScalarProducMatrix( mesh, nbeOfComponentsPrimal)
#h1ScalarProducMatrix = FT.ComputeH10ScalarProductMatrix(mesh, nbeOfComponentsPrimal)

##### ALGO (full) GREEDY
rb=greedy_algorithm(snapshots,TF,l2ScalarProducMatrix,None,1e-6,nev)

############################################################
"""          Rectification                               """
############################################################
reducedOrderBasisU=rb[0]

RI=np.zeros((TF,nev,nev)) #global rectification over the time steps
for time in range(TF):

    alpha=np.zeros((nev,ns)) #fine coefficients
    beta=np.zeros((ns,nev)) #coarse coefficients
    betainv=np.zeros((nev,ns)) 
    R=np.zeros((nev,nev)) #rectification matrix

    for j,elt in enumerate(parameters):

        u1PT = snapshots[j][time]
        u1T = snapshotsH[j][time]
        
        for i in range(nev):
            alpha[i,j]=u1PT@(l2ScalarProducMatrix@reducedOrderBasisU[i,:])
            beta[j,i]=u1T@(l2ScalarProducMatrix@reducedOrderBasisU[i,:])
            betainv[i,j]=beta[j,i]
        
    lambd=1e-13
    Rns=np.zeros((nev,ns))
    
    for j in range(ns):
        Rns[:,j]=np.linalg.inv(beta.transpose()@beta+lambd*np.eye(nev))@betainv[:,j]
        
    for i in range(nev):
        R[i,:]=Rns@alpha[i,:]

    RI[time,:,:]=R

############################################################
"""          Online part                                 """
############################################################
for i in [onlineParameter]:
    
    for cpt in range(nev,nev+1): #may add more test
        reducedOrderBasisU=rb[0]
       
        snapshotsHTime=[]
        
        for time in range(0,TG):
            snapshotH =MR.VTKReadToNp("Velocity",CoarsedataFolder+"/"+i+"/Snapshoth_",time)
   
            #Compute the projected data using the projection operator
            snapshotHSpaceinterpolated = snapshotHSpaceinterpolated = griddata(inputnodes,snapshotH,outputnodes,method='linear')#//Operator.dot(snapshotH)
            snapshotsHTime.append(snapshotHSpaceinterpolated)
    
        interp  = interp1d(oldtime,snapshotsHTime,kind='quadratic',axis=0,fill_value="extrapolate")
        solutionUHI=interp(newtime) #time and space interpolation
        
        for time in range(0,TF):
        
            R=RI[time,:,:]
            u1PT=solutionUHI[time]
            #u1PT=MR.VTKReadToNp("Velocity",FinedataFolder+"/"+str(i)+"/Snapshoth_",time) #true solution
            coef=np.zeros(cpt)
            CompressedSolutionUj=np.zeros(cpt)
            for k in range(cpt):
                
                coef[k]=0
                #CompressedSolutionUj[k]=u1PT@(l2ScalarProducMatrix@reducedOrderBasisU[k,:]) #traditional NIRB     
                for j in range(cpt):
                    coef[k]+=R[k,j]*u1PT@(l2ScalarProducMatrix@reducedOrderBasisU[j,:])
                #print(coef[k])
            
            reconstructedCompressedSolution = np.dot(coef, reducedOrderBasisU) #rectified nirb
            #reconstructedCompressedSolution = np.dot(CompressedSolutionUj, reducedOrderBasisUj) # traditional nirb
        
            ##################################################
            #######   saving solution in VTK ############
            VTKBase = MR.VTKReadmesh(meshFileName)
            MR.numpyToVTKWrite(VTKBase,reconstructedCompressedSolution,"NIRB_approximation_"+str(time)+"_"+str(cpt)+".vtu")
            ##################################################

        
